module ::Voltronic
  module Protocol
    require 'voltronic/digest'

  ##
  # Open a connection to the device
  # - Closes an existing connection, if one exists

    ##
    # Create a new Protocol object using an input IO object
    def self.for_io(io)
      require 'voltronic/protocols/io'

      def self.for_io(io)
        ::Voltronic::Protocol::IO.method(:new).call(io)
      end

      for_io(io)
    end

    ##
    # Create a new Protocol object for Axpert device connected to a serial port
    def self.for_serialport(port_or_dev, baud = 2400, data_bits = 8, stop_bits = 1, parity = :none)
      begin
        require 'serialport'
      rescue Exception
        raise LoadError.new 'RubyGem "serialport" required to make use of Axpert on a Serial Port'
      end

      def self.for_serialport(port_or_dev, baud = 2400, data_bits = 8, stop_bits = 1, parity = :none)
        port_or_dev = port_or_dev.to_s.strip

        baud = begin
          parse = Integer(baud)
          raise if (1 > parse)
          parse
        rescue
          raise ArgumentError.new "Invalid baud #{baud}"
        end

        data_bits = begin
          parse = Integer(data_bits)
          raise unless ((5 == parse) || (6 == parse) || (7 == parse) || (8 == parse))
          parse
        rescue
          raise ArgumentError.new "Invalid data bits #{data_bits}"
        end

        stop_bits = begin
          parse = Integer(stop_bits)
          raise unless ((1 == parse) || (2 == parse))
          parse
        rescue
          raise ArgumentError.new "Invalid stop bits #{stop_bits}"
        end

        parity = begin
          parse = parity.to_s.strip.upcase
          raise unless (('NONE' == parse) || ('EVEN' == parse) || ('ODD' == parse) || ('MARK' == parse) || ('SPACE' == parse))
          ::SerialPort.const_get(parse.to_sym)
        rescue
          raise ArgumentError.new "Invalid Parity #{parity}"
        end

        serial_io = ::SerialPort.new(port_or_dev, baud, data_bits, stop_bits, parity)
        serial_io.read_timeout = -1
        for_io(serial_io)
      end

      for_serialport(port_or_dev, baud, data_bits, stop_bits, parity)
    end

    ##
    # Create a new Protocol object for Axpert device connected to USB
    #
    # Currently only supported on Linux kernels that include HIDRaw
    def self.for_usb(dev)
      if !(RUBY_PLATFORM =~ /linux/)
        raise NotImplementedError.new 'USB is currently only supported in Linux'
      end

      def self.for_usb(dev)
        for_io(::File.open(dev.to_s, (::IO::RDWR | ::IO::NONBLOCK)))
      end

      for_usb(dev)
    end

    ##
    # Execute a query on the device
    def execute(input, timeout = 1.0)
      raise NotImplementedError.new 'Protocol implementation does not implement execute(..)'
    end
  end
end
