require 'voltronic/protocol'

##
# Implementation of the Protocol for IO objects
class ::Voltronic::Protocol::IO #:nodoc:
  require 'time'

  MAXIMUM_RESPONSE_SIZE = 1024
  READ_BLOCK_SIZE = 8 # USB device report size
  READ_OPTIONS = { exception: false }.freeze
  CLEAR_TIMEOUT = 2.0

  def initialize(io) #:nodoc:
    @io_array = [io].freeze

    @io = begin # Verify input io quacks like an IO
      begin
        io.sync = true
      rescue Exception
      end

      begin
        io.binmode = true
      rescue Exception
      end

      begin
        io.read_nonblock(READ_BLOCK_SIZE, nil, READ_OPTIONS)
      rescue Exception
        raise ArgumentError.new [io.to_s, ' does not support read_nonblock'].join
      end

      if !io.respond_to?(:write)
        raise ArgumentError.new [io.to_s, ' does not support write'].join
      end

      begin
        ::IO.select(@io_array, nil, nil, 0)
      rescue Exception
        raise ArgumentError.new [io.to_s, ' does not support IO.select(', io.to_s, ')'].join
      end

      io
    end

    @support_flush ||= begin # Assert if flush works
      @io.flush
      true
    rescue Exception
      false
    end

    clear_read_buffer
  end

  def execute(input, timeout = 1.0) #:nodoc:
    timeout = begin
      parse = Float(timeout)
      raise if (0 >= parse)
      parse
    rescue
      raise ArgumentError.new ['Invalid timeout ', timeout.to_s].join
    end

    clear_read_buffer
    write(input)
    read(timeout)
  end

  def to_s #:nodoc:
    'Protocol(IO)'
  end

  def inspect #:nodoc:
    self.to_s
  end

  def close
    @io.close
    true
  rescue Exception
    false
  end

  def closed?
    @io.closed?
  end

  private

  def clear_read_buffer #:nodoc:
    loop_timeout = Time.now.to_f + CLEAR_TIMEOUT

    while(true)
      if ::IO.select(@io_array, nil, nil, 0).nil?
        return true
      end

      @io.read_nonblock(READ_BLOCK_SIZE, nil, READ_OPTIONS)

      if (Time.now.to_f > loop_timeout)
        raise DeviceError.new 'Device is responding with binary data before a write has been issued'
      end
    end
  end

  def read(timeout) #:nodoc:
    blocks = ''
    loop_timeout = Time.now.to_f + timeout
    while(true)
      iteration_timeout = (loop_timeout - Time.now.to_f)
      if (0 >= iteration_timeout)
        raise TimeoutError.new ['Timeout of ', timeout.to_s, ' seconds reached'].join
      end

      next if ::IO.select(@io_array, nil, nil, iteration_timeout).nil?

      block = @io.read_nonblock(READ_BLOCK_SIZE, nil, READ_OPTIONS)
      next if block.nil?
      blocks << block

      if ::Voltronic::Digest.eos?(block)
        return ::Voltronic::Digest.parse(blocks)
      end

      if (blocks.length > MAXIMUM_RESPONSE_SIZE)
        raise BufferOverflowError.new ['Device response exceeds the maximum of ', MAXIMUM_RESPONSE_SIZE.to_s, ' bytes'].join
      end
    end
  end

  def write(input) #:nodoc:
    number_of_bytes = @io.write(::Voltronic::Digest.encode(input))
    if @support_flush
      @io.flush
    end
    number_of_bytes
  end

  class TimeoutError < RuntimeError; end #:nodoc:
  class BufferOverflowError < RuntimeError; end #:nodoc:
  class DeviceError < RuntimeError; end #:nodoc:

  private_class_method :new
end
